@@ -12,17 +12,14 @@
 
 #define LOG_TIMES	(1 << 0)
 
-void message(int level, char *format, ...)
+void message(int level, const char *format, ...)
 #ifdef __GNUC__
     __attribute__ ((format(printf, 2, 3)));
 #else
 ;
 #endif
-void logSetErrorFile(FILE * f);
 void logSetMessageFile(FILE * f);
 void logToSyslog(int enable);
-void logSetFlags(int flags);
-void logClearFlags(int flags);
 void logSetLevel(int level);
 
 #endif
